# This file is part of SSHare.
#
# SSHare is free software: you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# SSHare is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# SSHare. If not, see <https://www.gnu.org/licenses/>.

import argparse
import getpass
import os
import os.path
import time
import tomllib
import subprocess
import sys
from pathlib import Path

from sshare.logger         import Logger
from sshare.plugin         import Plugin
from sshare.plugins.config import Flag
from sshare.plugins.config import NoArgument
from sshare.plugins.config import NoDefault
from sshare.version        import version


def main():
    config_directory = Path(os.environ.get("XDG_CONFIG_DIR") or f"{os.environ["HOME"]}/.config") / "sshare"
    with open(config_directory / "config.toml", mode="rb") as file:
        config = tomllib.load(file)

    # Load command line early and set it as the active logger
    # so that it can be used to report errors while loading and
    # configuring plugins
    # i.e. before other logging plugins have had a chance to be initialised
    logger = Logger(preload_command_line=True)

    # Load inbuilt plugins
    plugins_flat = [
        Plugin.internal("file"),
        Plugin.internal("current_time"),
        Plugin.internal("append_type"),
        Plugin.internal("ssh"),
        Plugin.internal("log_result"),
    ]
    plugins = {}
    for type in [ "logger", "source", "name", "upload", "result" ]:
        plugins[type] = { "active": [], "inactive": [] }

    # Load external plugins
    sys.dont_write_bytecode = True
    for path in (config_directory / "plugins").iterdir():
        if path.is_file() and path.suffix == ".py":
            plugins_flat.append(Plugin.external(path))
    sys.dont_write_bytecode = False

    # Set plugin configurations from config file
    # Load plugin arguments and detect conflicts
    error = False
    argument_map = {}
    used_arguments = {}
    parser = argparse.ArgumentParser(
        prog = "sshare",
        description = "Upload files to a server via ssh",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s version {version}",
    )
    if config.get("plugins") == None:
        config["plugins"] = {}
    for plugin in plugins_flat:
        if hasattr(plugin, "config"):
            plugin_config = config["plugins"].get(plugin.name)
            if plugin_config != None:
                for config_entry in plugin_config.items():
                    plugin.config[config_entry[0]] = config_entry[1]
        else:
            setattr(plugin, "config", {})
        if hasattr(plugin, "args"):
            for arg_name, arg in plugin.args.items():
                if arg.is_valid():
                    arg.bind(plugin, arg_name)
                    def check_flag(flag):
                        if flag in used_arguments:
                            logger.error(f"Error: Argument '{arg_name}' for plugin '{plugin.name}' has conflict. Flag '{flag}' is also used by plugin '{used_arguments[arg.short]}'")
                            error = True
                    check_flag(arg.short)
                    check_flag(arg.long)
                    arg.add(parser, used_arguments)
                    argument_map[arg.dest] = plugin, arg_name
                else:
                    logger.error(f"Error: Argument '{arg_name}' must set either one or both of short and long flag parameters")
                    error = True
    if error:
        sys.exit(1)

    arguments = parser.parse_args()
    for arg, (plugin, config) in list(argument_map.items()):
        value = getattr(arguments, arg)
        if value != NoArgument:
            if value != Flag:
                plugin.config[config] = value
            del argument_map[arg]

    # Sort plugins by type and check activation criteria
    error = False
    for plugin in plugins_flat:
        if isinstance(plugin.plugin_type, str):
            plugin.plugin_type = [ plugin.plugin_type ]
        for plugin_type in plugin.plugin_type:
            plugins_of_type = plugins.get(plugin_type)
            if plugins_of_type == None:
                logger.error(f"Error: Plugin '{plugin.name}' has an invalid plugin type '{plugin_type}'")
                error = True
            else:
                active = True
                if hasattr(plugin, "activate"):
                    criteria = plugin.activate
                    if isinstance(plugin.activate, dict):
                        criteria = plugin.activate.get(plugin_type)
                    if criteria != None:
                        for criterion in criteria:
                            active = not plugin.args[criterion].dest in argument_map
                            if not active:
                                break
                plugins_of_type["active" if active else "inactive"].append(plugin)
                if active:
                    for config_entry in plugin.config.items():
                        if config_entry[1] == NoDefault:
                            logger.error(f"Error: Value 'plugins.{plugin.name}.{config_entry[0]}' has no default value and must be specified explicitly")
                            error = True
    for plugin_type, plugins_of_type in plugins.items():
        if len(plugins_of_type["active"]) == 0 and plugin_type != "logger":
            if len(plugins_of_type["inactive"]) == 0:
                logger.error(f"No '{plugin_type}' plugins available. Atleast one must be provided")
            else:
                logger.error(f"No '{plugin_type}' plugins activated. Activate at least one of:")
                for plugin in plugins_of_type["inactive"]:
                    logger.error(f"{plugin.name}:")
                    criteria = plugin.activate
                    if isinstance(plugin.activate, dict):
                        criteria = plugin.activate[plugin_type]
                    for criterion in criteria:
                        logger.error(f"  {plugin.args[criterion].pretty()}")
            error = True
    if error:
        sys.exit(1)

    # Objectify configs
    error = False
    class PluginConfig: pass
    for plugin in plugins_flat:
        config = plugin.config
        plugin.config = PluginConfig()
        for config_entry in config.items():
            setattr(plugin.config, config_entry[0], config_entry[1])
    if error:
        sys.exit(1)

    # Initialise plugins
    for plugin in plugins_flat:
        setattr(plugin, "logger", logger)
        if hasattr(plugin, "init"):
            error = error or plugin.init()

    logger.add(*plugins["logger"]["active"])

    sources = []
    for plugin in plugins["source"]["active"]:
        sources.append(plugin.get_source())
    if len(sources) == 0:
        logger.error("Error: No sources provided. Must activate at least one source plugin")
        log_activations(logger, plugins["source"])

    for index, source in enumerate(sources):
        name = ""
        for plugin in plugins["name"]["active"]:
            name = plugin.get_name(name, source)
        sources[index] = name, source

    for name, source in sources:
        for plugin in plugins["upload"]["active"]:
            plugin.upload(name, source)

    for name, _ in sources:
        for plugin in plugins["result"]["active"]:
            plugin.result(name)

    sys.exit(0)
